import os
import tempfile
from typing import Any, Dict

import pytest

from templafirm.core.templater import Templater


@pytest.fixture
def gke_test_templater() -> Templater:
    templater = Templater()
    return templater


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_template_str, template_inputs, templated_resource",
    [
        (
            'module "mrmda_node_pool" {\n  source = "../../modules/mrdma_node_pool/"\n\n  account_id                        = "test-project"\n  autoscaling                       = {\n    total_min_node_count = "0"\n    total_max_node_count = "4"\n  }\n  cluster_name                      = "test_cluster"\n  disk_size                         = "100"\n  disk_type                         = "fast"\n  ephemeral_storage_local_ssd_count = "0"\n  gpu_accelerator                   = {\n    count = "1"\n    type  = "h200"\n  }\n  image_type                        = "image_type"\n  labels                            = {\'label\': \'label\'}\n  machine_type                      = "a3-ultragpu-8g"\n  node_pool_name                    = "test-node-pool" \n  node_region                       = "us-central1"\n  node_sa_email                     = "somebody@email.com" \n  node_zone                         = "b"\n  placement_policy                  = {\n    type = "COMPACT"\n  }\n  reservation_affinity              = {\n    type = "SPECIFIC_RESERVATION"\n    reservations = ["test_id"]\n  }\n}',
            {
                "cluster_name": "test_cluster",
                "disk_size": "100",
                "dist_type": "fast",
                "ephemeral_storage_local_ssd_count": "0",
                "gcp_project_id": "test-project",
                "gpu_accelerator_count": "1",
                "gpu_accelerator_type": "h200",
                "image_type": "image_type",
                "labels": {"label": "label"},
                "machine_type": "a3-ultragpu-8g",
                "node_pool_name": "test-node-pool",
                "node_region": "us-central1",
                "node_sa_email": "somebody@email.com",
                "node_zone": "b",
                "placement_policy_type": "COMPACT",
                "reservation_ids": '["test_id"]',
                "reservation_type": "SPECIFIC_RESERVATION",
                "total_max_node_count": "4",
                "total_min_node_count": "0",
            },
            "mrdma_node_pool",
        ),
    ],
)
async def test_gke_templater_generation(
    gke_test_templater: Templater,
    expected_template_str: str,
    template_inputs: Dict[str, Any],
    templated_resource: str,
) -> None:
    provider_and_env = gke_test_templater.return_provider(provider_key="gke")
    assert templated_resource in provider_and_env.provider._provider_meta_table.template_mapping.keys()

    gke_test_templater.activate_provider("gke")
    templated_resource_rendered = await gke_test_templater.render_template_resource(
        template_resource_inputs=template_inputs, template_resource_name=templated_resource
    )
    assert templated_resource_rendered == expected_template_str


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_template_str, template_inputs, templated_resource",
    [
        (
            'module "mrmda_node_pool" {\n  source = "../../modules/mrdma_node_pool/"\n\n  account_id                        = "test-project"\n  autoscaling                       = {\n    total_min_node_count = "0"\n    total_max_node_count = "4"\n  }\n  cluster_name                      = "test_cluster"\n  disk_size                         = "100"\n  disk_type                         = "fast"\n  ephemeral_storage_local_ssd_count = "0"\n  gpu_accelerator                   = {\n    count = "1"\n    type  = "h200"\n  }\n  image_type                        = "image_type"\n  labels                            = {\'label\': \'label\'}\n  machine_type                      = "a3-ultragpu-8g"\n  node_pool_name                    = "test-node-pool" \n  node_region                       = "us-central1"\n  node_sa_email                     = "somebody@email.com" \n  node_zone                         = "b"\n  placement_policy                  = {\n    type = "COMPACT"\n  }\n  reservation_affinity              = {\n    type = "SPECIFIC_RESERVATION"\n    reservations = ["test_id"]\n  }\n}',
            {
                "cluster_name": "test_cluster",
                "disk_size": "100",
                "dist_type": "fast",
                "ephemeral_storage_local_ssd_count": "0",
                "gcp_project_id": "test-project",
                "gpu_accelerator_count": "1",
                "gpu_accelerator_type": "h200",
                "image_type": "image_type",
                "labels": {"label": "label"},
                "machine_type": "a3-ultragpu-8g",
                "node_pool_name": "test-node-pool",
                "node_region": "us-central1",
                "node_sa_email": "somebody@email.com",
                "node_zone": "b",
                "placement_policy_type": "COMPACT",
                "reservation_ids": '["test_id"]',
                "reservation_type": "SPECIFIC_RESERVATION",
                "total_max_node_count": "4",
                "total_min_node_count": "0",
            },
            "mrdma_node_pool",
        ),
    ],
)
async def test_gke_templater_file_generation(
    gke_test_templater: Templater,
    expected_template_str: str,
    template_inputs: Dict[str, Any],
    templated_resource: str,
) -> None:
    provider_and_env = gke_test_templater.return_provider(provider_key="gke")
    assert templated_resource in provider_and_env.provider._provider_meta_table.template_mapping.keys()

    gke_test_templater.activate_provider("gke")

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = os.path.join(tmpdir, "test_template.tf")
        gke_test_templater.activate_provider("gke")
        await gke_test_templater.render_template_resource_to_file(
            output_path=output_path, template_resource_inputs=template_inputs, template_resource_name=templated_resource
        )

        assert os.path.exists(output_path)

        with open(output_path, "r") as open_template_buffer:
            templated_resource = "".join(open_template_buffer.readlines())
        assert templated_resource == expected_template_str



import os
from collections.abc import AsyncGenerator

import pytest
import pytest_asyncio
import tftest

from templafirm.core.templater import Templater


@pytest_asyncio.fixture
async def plan() -> AsyncGenerator:
    # purposefully do not use a temp dir as it borks tf import paths
    templater = Templater()

    template_resource_name = "mrdma_node_pool"
    template_inputs = {
        "cluster_name": "test_cluster",
        "disk_size": "100",
        "dist_type": "fast",
        "ephemeral_storage_local_ssd_count": "0",
        "gcp_project_id": "test-project",
        "gpu_accelerator_count": "1",
        "gpu_accelerator_type": "h200",
        "image_type": "image_type",
        "labels": '{"label": "label"}',
        "machine_type": "a3-ultragpu-8g",
        "node_pool_name": "test-node-pool",
        "node_region": "us-central1",
        "node_sa_email": "somebody@email.com",
        "node_zone": "b",
        "placement_policy_type": "COMPACT",
        "reservation_ids": '["test-id"]',
        "reservation_type": "SPECIFIC_RESERVATION",
        "total_max_node_count": "4",
        "total_min_node_count": "0",
    }

    node_pool_dir = os.path.join(templater.return_provider("gke").provider.template_directory_path(), "node_pools")
    output_path = os.path.join(node_pool_dir, "test_template.tf")
    templater.activate_provider("gke")
    await templater.render_template_resource_to_file(
        output_path=output_path, template_resource_inputs=template_inputs, template_resource_name=template_resource_name
    )

    assert os.path.exists(output_path)

    tf_tester = tftest.TerraformTest(node_pool_dir, enable_cache=True)
    tf_tester.setup()
    tf_tester.init(input=True)
    yield tf_tester.plan(output=True)
    os.remove(output_path)


@pytest.mark.asyncio
async def test_gvnic_plan(plan: tftest.TerraformPlanOutput) -> None:
    expected_net_set = {
        "module.mrmda_node_pool.google_compute_network.gvnic_mrdma_vpc",
        "module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_gvnic_mrdma",
    }
    resource_change_key_set = set(plan.resource_changes.keys())
    assert expected_net_set.intersection(resource_change_key_set) == expected_net_set

    gvnic_net_change = plan.resource_changes["module.mrmda_node_pool.google_compute_network.gvnic_mrdma_vpc"]
    assert gvnic_net_change["change"]["after"]["name"] == "a3-ultragpu-8g-us-central1-b-test-id-gvnic"

    gvnic_subnet_change = plan.resource_changes[
        "module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_gvnic_mrdma"
    ]["change"]["after"]
    assert gvnic_subnet_change["ip_cidr_range"] == "192.170.1.0/24"
    assert gvnic_subnet_change["name"] == "gvnic-sub-a3-ultragpu-8g-us-central1-b-test-id"
    assert gvnic_subnet_change["project"] == "test-project"
    assert gvnic_subnet_change["region"] == "us-central1"


@pytest.mark.asyncio
async def test_mrdma_plan(plan: tftest.TerraformPlanOutput) -> None:
    expected_subnet_set = {
        "module.mrmda_node_pool.google_compute_network.vpc_gke_roce",
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-0"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-1"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-2"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-3"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-4"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-5"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-6"]',
        'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-7"]',
    }
    resource_change_key_set = set(plan.resource_changes.keys())
    assert expected_subnet_set.intersection(resource_change_key_set) == expected_subnet_set

    # test the name of the gke_roce net
    mrdma_roce_change = plan.resource_changes["module.mrmda_node_pool.google_compute_network.vpc_gke_roce"]
    assert mrdma_roce_change["change"]["after"]["name"] == "a3-ultragpu-8g-us-central1-b-test-id-mrdma"

    # test data elements
    for i in range(0, 8):
        mrdma_subnet_change = plan.resource_changes[
            f'module.mrmda_node_pool.google_compute_subnetwork.subnet_gke_roce["a3-ultragpu-8g-us-central1-b-test-id-{i}"]'
        ]["change"]["after"]
        assert (
            mrdma_subnet_change["description"]
            == f"The {i}th subnet for GPUDirect MRDMA network a3-ultragpu-8g-us-central1-b-test-id-mrdma"
        )
        assert mrdma_subnet_change["ip_cidr_range"] == f"192.169.{i}.0/24"
        assert mrdma_subnet_change["name"] == f"roce-sub-a3-ultragpu-8g-us-central1-b-test-id-{i}"
        assert mrdma_subnet_change["project"] == "test-project"
        assert mrdma_subnet_change["region"] == "us-central1"


@pytest.mark.asyncio
async def test_node_pool_plan(plan: tftest.TerraformPlanOutput) -> None:
    expected_node_pool_set = {"module.mrmda_node_pool.google_container_node_pool.gpu_mrdma_node_pool"}
    resource_change_key_set = set(plan.resource_changes.keys())
    assert expected_node_pool_set.intersection(resource_change_key_set) == expected_node_pool_set

    node_pool_change = plan.resource_changes["module.mrmda_node_pool.google_container_node_pool.gpu_mrdma_node_pool"][
        "change"
    ]["after"]

    # test expected resource_changes

    assert node_pool_change["autoscaling"] == [
        {"max_node_count": None, "min_node_count": None, "total_max_node_count": 4, "total_min_node_count": 0}
    ]
    assert node_pool_change["cluster"] == "test_cluster"
    assert node_pool_change["location"] == "us-central1"
    assert node_pool_change["network_config"] == [
        {
            "additional_node_network_configs": [
                {"subnetwork": "gvnic-sub-a3-ultragpu-8g-us-central1-b-test-id"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-0"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-1"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-2"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-3"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-4"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-5"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-6"},
                {"subnetwork": "roce-sub-a3-ultragpu-8g-us-central1-b-test-id-7"},
            ],
            "additional_pod_network_configs": [],
            "create_pod_range": None,
            "enable_private_nodes": True,
            "network_performance_config": [],
        }
    ]
    assert node_pool_change["node_config"] == [
        {
            "advanced_machine_features": [],
            "boot_disk_kms_key": None,
            "containerd_config": [],
            "disk_size_gb": 100,
            "disk_type": "fast",
            "enable_confidential_storage": None,
            "ephemeral_storage_local_ssd_config": [{"data_cache_count": None, "local_ssd_count": 0}],
            "fast_socket": [{"enabled": False}],
            "guest_accelerator": [
                {
                    "count": 1,
                    "gpu_driver_installation_config": [{"gpu_driver_version": "LATEST"}],
                    "gpu_partition_size": None,
                    "gpu_sharing_config": [],
                    "type": "h200",
                }
            ],
            "gvnic": [{"enabled": True}],
            "host_maintenance_policy": [],
            "image_type": "image_type",
            "labels": {"label": "label"},
            "linux_node_config": [],
            "local_nvme_ssd_block_config": [],
            "local_ssd_encryption_mode": None,
            "machine_type": "a3-ultragpu-8g",
            "max_run_duration": None,
            "metadata": {"disable-legacy-endpoints": "true"},
            "node_group": None,
            "oauth_scopes": [
                "https://www.googleapis.com/auth/devstorage.read_only",
                "https://www.googleapis.com/auth/logging.write",
                "https://www.googleapis.com/auth/monitoring",
                "https://www.googleapis.com/auth/service.management.readonly",
                "https://www.googleapis.com/auth/servicecontrol",
                "https://www.googleapis.com/auth/trace.append",
            ],
            "preemptible": False,
            "reservation_affinity": [
                {
                    "consume_reservation_type": "SPECIFIC_RESERVATION",
                    "key": "compute.googleapis.com/reservation-name",
                    "values": ["test-id"],
                }
            ],
            "resource_labels": None,
            "resource_manager_tags": None,
            "secondary_boot_disks": [],
            "service_account": "somebody@email.com",
            "sole_tenant_config": [],
            "spot": False,
            "storage_pools": None,
            "tags": ["test-id"],
            "taint": [],
        }
    ]


@pytest.mark.asyncio
async def test_services_plan(plan: tftest.TerraformPlanOutput) -> None:
    expected_services_set = {
        "module.mrmda_node_pool.google_project_service.usage_service",
        "module.mrmda_node_pool.google_project_service.k8s_service",
        "module.mrmda_node_pool.google_project_service.compute_service",
    }
    resource_change_key_set = set(plan.resource_changes.keys())
    assert expected_services_set.intersection(resource_change_key_set) == expected_services_set



import os
import pathlib
from typing import Any, Dict

import pytest
import yaml

from templafirm.core.meta_table import ProviderMetaTable, ResourceTemplate
from templafirm.core.provider import Provider
from test import assets

TEST_PROVIDER_META_PATH = "test/assets/test_provider_meta.yaml"


class BaseProviderTester(Provider):
    def __init__(self, provider_meta_path: str):
        super().__init__(provider_meta_path)

    def template_directory_path(self) -> str:
        template_module_file_path = pathlib.Path(os.path.abspath(assets.__file__))
        template_module_file_path = pathlib.Path(os.path.join(os.getcwd(), template_module_file_path))
        return str(template_module_file_path.parent)


@pytest.fixture
def base_test_provider() -> BaseProviderTester:
    return BaseProviderTester(TEST_PROVIDER_META_PATH)


@pytest.fixture
def base_test_provider_raw_yaml() -> Dict[str, Any]:
    with open(TEST_PROVIDER_META_PATH, "r") as open_meta_yaml_buffer:
        yaml_data = yaml.safe_load(open_meta_yaml_buffer)
        return yaml_data


def test_base_provider_initialization(
    base_test_provider: Provider, base_test_provider_raw_yaml: Dict[str, Any]
) -> None:
    expected_template_object = ProviderMetaTable(
        name=base_test_provider_raw_yaml["name"],
        version=base_test_provider_raw_yaml["version"],
        template_mapping={
            "test_empty_resource_template": ResourceTemplate(
                name=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"]["name"],
                version=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"]["version"],
                template_file_path=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"][
                    "template_file_path"
                ],
            ),
            "test_no_input_resource_template": ResourceTemplate(
                name=base_test_provider_raw_yaml["template_mapping"]["test_no_input_resource_template"]["name"],
                version=base_test_provider_raw_yaml["template_mapping"]["test_no_input_resource_template"]["version"],
                template_file_path=base_test_provider_raw_yaml["template_mapping"]["test_no_input_resource_template"][
                    "template_file_path"
                ],
            ),
            "test_input_resource_template": ResourceTemplate(
                name=base_test_provider_raw_yaml["template_mapping"]["test_input_resource_template"]["name"],
                version=base_test_provider_raw_yaml["template_mapping"]["test_input_resource_template"]["version"],
                template_file_path=base_test_provider_raw_yaml["template_mapping"]["test_input_resource_template"][
                    "template_file_path"
                ],
            ),
        },
    )
    assert base_test_provider._provider_meta_table == expected_template_object


def test_base_provider_resources(base_test_provider: Provider, base_test_provider_raw_yaml: Dict[str, Any]) -> None:
    expected_resource_keys = base_test_provider_raw_yaml["template_mapping"].keys()
    assert base_test_provider.resources == expected_resource_keys


def test_base_provider_version(base_test_provider: Provider, base_test_provider_raw_yaml: Dict[str, Any]) -> None:
    assert base_test_provider.version == base_test_provider_raw_yaml["version"]


def test_base_provider_resource_get(base_test_provider: Provider, base_test_provider_raw_yaml: Dict[str, Any]) -> None:
    expected_resource = ResourceTemplate(
        name=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"]["name"],
        version=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"]["version"],
        template_file_path=base_test_provider_raw_yaml["template_mapping"]["test_empty_resource_template"][
            "template_file_path"
        ],
    )
    assert base_test_provider["test_empty_resource_template"] == expected_resource



import pytest

from templafirm.core.meta_table import ResourceTemplate
from templafirm.gke.gke_provider import GKEProvider

EXPECTED_PROVIDER_VERSION = "0.0.1"
EXPECTED_PROVIDER_NAME = "gke"


@pytest.fixture
def gke_test_provider() -> GKEProvider:
    return GKEProvider()


def test_gke_provider_initialization(gke_test_provider: GKEProvider) -> None:
    assert gke_test_provider.version == EXPECTED_PROVIDER_VERSION
    assert gke_test_provider.name == EXPECTED_PROVIDER_NAME


def test_gke_provider_mrdma_node_pool_resource(gke_test_provider: GKEProvider) -> None:
    expected_mrdma_node_pool_resource = ResourceTemplate(
        name="mrdma_node_pool",
        version="0.0.1",
        template_file_path="node_pools/mrdma_node_pool.jinja2",
        template_inputs={
            "cluster_name",
            "disk_size",
            "dist_type",
            "ephemeral_storage_local_ssd_count",
            "gcp_project_id",
            "gpu_accelerator_count",
            "gpu_accelerator_type",
            "image_type",
            "labels",
            "machine_type",
            "node_pool_name",
            "node_region",
            "node_sa_email",
            "node_zone",
            "placement_policy_type",
            "reservation_ids",
            "reservation_type",
            "total_max_node_count",
            "total_min_node_count",
        },
    )
    assert expected_mrdma_node_pool_resource == gke_test_provider["mrdma_node_pool"]



import dataclasses
from typing import Dict, Set


@dataclasses.dataclass
class ResourceTemplate:
    """DC for resource to be templated.

    Args:
        description (str, optional): Description of the resource, defaults to ''.
        file_extension (str, optional): Type of file, defaults to '.tf'.
        name (str): Name of the resource.
        version (str): Version of the resource.
        template_file_path (str): Path in the template directory to template file.
        template_inputs (Set[str]): Inputs into the templated resource, defaults to empty set.

    Attributes:
        description (str): Description of the resource, defaults to ''.
        file_extension (str): Type of file, defaults to '.tf'.
        name (str): Name of the resource.
        version (str): Version of the resource.
        template_file_path (str): Path in the template directory to template file.
        template_inputs (Set[str]): Inputs into the templated resource, defaults to empty set.

    """

    name: str
    version: str
    template_file_path: str
    description: str = ""
    file_extension: str = ".tf"
    template_inputs: Set[str] = dataclasses.field(default_factory=set)


@dataclasses.dataclass
class ProviderMetaTable:
    """DC for entire resource provider.

    Args:
        description (str, optional): Description of the provider, defaults to ''.
        name (str): Name of the provider.
        version (str): Version of the provider.
        template_mapping (Dict[str, ResourceTemplate]): Mapping of template name to definition.

    Attributes:
        description (str, optional): Description of the provider, defaults to ''.
        name (str): Name of the provider.
        version (str): Version of the provider.
        template_mapping (Dict[str, ResourceTemplate]): Mapping of template name to definition.
    """

    name: str
    version: str
    template_mapping: Dict[str, ResourceTemplate]
    description: str = ""

    def __contains__(self, resource_name: str) -> bool:
        return resource_name in self.template_mapping

    def __getitem__(self, resource_key: str) -> ResourceTemplate:
        if resource_key not in self.template_mapping:
            raise KeyError(f"{resource_key} not registered in provider.")

        return self.template_mapping[resource_key]



import os
from abc import ABC, abstractmethod
from collections.abc import KeysView

import yaml

from templafirm.core.meta_table import ProviderMetaTable, ResourceTemplate


class Provider(ABC):
    """Base class struct for providing template structs to Jinja engine.


    Args:
        provider_meta_path (str): Path to the meta data table.

    Attributes:
        _provider_meta_path (str): Path to the meta data table.
        _provider_meta_table (ProviderMetaTable): Meta table object storing provider meta and resources.
    """

    def __init__(self, provider_meta_path: str):
        self._provider_meta_path = provider_meta_path
        self._provider_meta_table = self._load_meta_table()

    @abstractmethod
    def template_directory_path(self) -> str:
        """Get the defined path to the templates.

        Raises:
            NotImplementedError: When not implemented in subclass.

        Returns:
            str: The path on the system to the template directory.
        """
        raise NotImplementedError(
            "Need to implement method to locate your template directory on fs. Done for Jinja2 FileSystemLoader."
        )

    def __verify_meta_table(self, provider_meta_table: ProviderMetaTable) -> None:
        """Verify the template files that are present in the meta table."""
        for _, resource_template in provider_meta_table.template_mapping.items():
            resource_template_path = os.path.join(
                os.path.dirname(self._provider_meta_path), resource_template.template_file_path
            )
            assert os.path.exists(resource_template_path)

    def _load_meta_table(self) -> ProviderMetaTable:
        """Load the meta table.

        Returns:
            ProviderMetaTable: Meta table for the provider containing the provider data and template data.
        """
        with open(self._provider_meta_path, "r") as open_provider_meta_buffer:
            provider_yaml_obj = yaml.safe_load(open_provider_meta_buffer)
            provider_meta_table = ProviderMetaTable(**provider_yaml_obj)
            # need to parse the resources individually as dataclass constructor doesn't form in sub class objects
            # these are silently type failed as they come across as dicts.
            resource_yaml_defs = provider_meta_table.template_mapping
            resource_obj_defs = {}
            for resource_name, resource_yaml_def in resource_yaml_defs.items():
                # type ignore on this because mypy thinks the type is ResourceTemplate but it gets loaded as dict
                resource_template_obj = ResourceTemplate(**resource_yaml_def)  # type: ignore
                # have to do the same thing to the set of inputs
                resource_template_obj.template_inputs = (
                    set(resource_yaml_def["template_inputs"]) if "template_inputs" in resource_yaml_def else set()  # type: ignore
                )
                resource_obj_defs[resource_name] = resource_template_obj
            provider_meta_table.template_mapping = resource_obj_defs
        # verify all files in the table exist in the described directory
        self.__verify_meta_table(provider_meta_table)
        return provider_meta_table

    @property
    def name(self) -> str:
        """Get name of provider.

        Returns:
            str: Provider name.
        """
        return self._provider_meta_table.name

    @property
    def version(self) -> str:
        """Get version of provider.

        Returns:
            str: Provider version.
        """
        return self._provider_meta_table.version

    @property
    def resources(self) -> KeysView:
        """Get resource names as KeysView.

        Returns:
            KeysView[str]: Keys for provider defined resources.
        """
        return self._provider_meta_table.template_mapping.keys()

    def __contains__(self, resource_name: str) -> bool:
        return resource_name in self._provider_meta_table

    def __getitem__(self, resource_name: str) -> ResourceTemplate:
        if resource_name not in self._provider_meta_table.template_mapping:
            raise KeyError(f"{resource_name} does not exist in {self.__class__.__name__}.")

        return self._provider_meta_table[resource_name]



import os
import pathlib

from templafirm.core.provider import Provider
from templafirm.gke import templates


class GKEProvider(Provider):
    """Template provider for GKE. Abstraction on top of file path provided in package.

    Attributes:
        _localized_assets_pathway (str): pathway in fs to the templates directory that is packaged
            with templafirm.
    """

    def __init__(self) -> None:
        self._localized_assets_pathway = self._get_localized_assets_path()
        super().__init__(self._localized_assets_pathway)

    def template_directory_path(self) -> str:
        """Return the template directory relative to python env.

        Returns:
            str: pathway to the template directory.
        """
        template_module_file_path = pathlib.Path(os.path.abspath(templates.__file__))
        return str(template_module_file_path.parent)

    def _get_localized_assets_path(self) -> str:
        """Get the localized assets path.

        Returns:
            str: pathway to local resource definitions.
        """
        gke_provider_meta_path = os.path.join(self.template_directory_path(), "gke_provider_meta.yaml")
        return gke_provider_meta_path



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
project = "templafirm"
copyright = "2025, Max Weirz"
author = "Max Weirz"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"

html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]



import asyncio
import logging
import os
from collections.abc import KeysView
from dataclasses import dataclass
from typing import Any, Dict, Optional

from jinja2 import Environment, FileSystemLoader

from templafirm.core.provider import Provider
from templafirm.gke.gke_provider import GKEProvider


@dataclass
class ProviderAndEnvironment:
    """Dataclass containing provider, associated jinja env, and lock for provider access.

    Args:
        environment (Environment): Jinja2 environment for template generation.
        provider (Provider): Provider object for defining templates.
        provider_lock (asyncio.Lock): Lock for provider access across threads.

    Attributes:
        environment (Environment): Jinja2 environment for template generation.
        provider (Provider): Provider object for defining templates.
        provider_lock (asyncio.Lock): Lock for provider access across threads.
    """

    environment: Environment
    provider: Provider
    provider_lock: asyncio.Lock


@dataclass
class ProviderRegistry:
    """Provider register dataclass.

    Holds the providers that are available as well as some easy access and setter methods.

    Attributes:
        providers (Dict[str, ProviderAndEnvironment]): Mapping from provider names to objects registered into registry.

    Raises:
        KeyError: Raises if provider key doesn't exist in registry.
    """

    providers: Dict[str, ProviderAndEnvironment]

    def keys(self) -> KeysView:
        """Get keys of providers available.

        Returns:
            KeysView[str]: Key view of available providers.
        """
        return self.providers.keys()

    def __getitem__(self, provider_key: str) -> ProviderAndEnvironment:
        """Get provider by key.

        Args:
            provider_key (str): Key of provider.

        Raises:
            KeyError: Raises if provider key is not registered.

        Returns:
            ProviderAndEnvironment: ProviderAndEnvironment obj corresponding to provider and associated jinja env.
        """
        if provider_key not in self.providers:
            raise KeyError(f"{provider_key} not a registered provider.")

        return self.providers[provider_key]

    def __setitem__(self, provider_key: str, provider_and_env: ProviderAndEnvironment) -> None:
        """Set provider in registry by key

        Args:
            provider_key (str): Key of provider.
            provider_and_env (ProviderAndEnvironment): ProviderAndEnvironment obj corresponding to provider and associated jinja env.
        """
        self.providers[provider_key] = provider_and_env

    def __contains__(self, provider_name: str) -> bool:
        """Determine if provider is registry.

        Args:
            provider_name (str): Key of provider in registry.

        Returns:
            bool: Whether provider exists in registry.
        """
        return provider_name in self.providers


class Templater:
    """Renders files from templates provider by registered providers.

    Supports adding new providers to the class level registry at runtime for custom
    provider support.

    Args:
        provider_lock_timeout (int, optional): Seconds to timeout on provider lock. Defaults to 5.

    Attributes:
        gke_provider (GKEProvider): Provider object for gke resources.
        _provider_lock_timeout (int): Time in seconds before provider lock timesout.
        _provider_registry (ProviderRegistry): Registry of providers available to all class instances.

    Raises:
        asyncio.TimeoutError: Raises if provider lock times out across instances. Done to ensure
            locking across all async Jinja envs.
        KeyError: Raises if user attempts access to un-registered provider or template.
    """

    gke_provider = GKEProvider()
    _provider_registry: ProviderRegistry = ProviderRegistry(
        providers={
            "gke": ProviderAndEnvironment(
                provider=gke_provider,
                provider_lock=asyncio.Lock(),
                environment=Environment(
                    loader=FileSystemLoader(searchpath=gke_provider.template_directory_path()), enable_async=True
                ),
            )
        }
    )

    def __init__(self, provider_lock_timeout: int = 5):
        self._active_provider = self._provider_registry["gke"]
        self._provider_lock_timeout = provider_lock_timeout

    @classmethod
    def register_provider(cls, provider_key: str, provider_object: Provider) -> None:
        """Register provider to class registry.

        Args:
            provider_key (str): Key of provider to register.
            provider_object (Provider): Provider object to register.
        """
        if provider_key in cls._provider_registry:
            logging.warning(f"Overwriting previous registration of provider {provider_key}.")
        cls._provider_registry[provider_key] = ProviderAndEnvironment(
            provider=provider_object,
            provider_lock=asyncio.Lock(),
            environment=Environment(
                loader=FileSystemLoader(searchpath=provider_object.template_directory_path()), enable_async=True
            ),
        )

    @classmethod
    def return_provider(cls, provider_key: str) -> ProviderAndEnvironment:
        """Return the ProviderAndEnvironment dc if available.

        Args:
            provider_key (str): The key of the provider to return.

        Returns:
            ProviderAndEnvironment: Provider and environment dataclass.
        """
        return cls._provider_registry[provider_key]

    @classmethod
    def list_providers(cls) -> KeysView:
        """List registered providers

        Returns:
            KeysView[str]: dict keys of the registered providers.
        """
        return cls._provider_registry.keys()

    @classmethod
    def list_templates(cls, provider_name: str) -> KeysView:
        """List available templates for chosen provider.

        Args:
            provider_name (str): Name of provider to list templates.

        Returns:
            KeysView[str]: Key view of registered templates in chosen provider.
        """
        return cls._provider_registry[provider_name].provider._provider_meta_table.template_mapping.keys()

    def __verify_output_file_type(self, output_path: str, template_resource_name: str) -> bool:
        """Verify output path has expected file type.

        Args:
            output_path (str): desired output path for template resource.
            template_resource_name (str): name of the resource to template generate.

        Returns:
            bool: whether the file type is of the right type.
        """
        extension = os.path.splitext(output_path)[1]
        # don't need to lock as only doing look up and not using attached environment
        expected_extension = self._active_provider.provider[template_resource_name].file_extension
        if expected_extension != extension:
            logging.error(
                "Didn't generate template, provided path does not have expected file type.\nExpected {self._active_provider[template_resource_name].file_extension}."
            )
        return expected_extension == extension

    def activate_provider(self, provider_name: str) -> None:
        """Set the active provider.

        Args:
            provider_name str: Name of the provider to set to active state.
        """
        self._active_provider = self._provider_registry[provider_name]

    async def acquire_provider_lock(self) -> asyncio.Lock:
        """Acquire the provider lock.

        Raises:
            asyncio.TimeoutError: Raises if provider locking times out.

        Returns:
            asyncio.Lock: Async lock object to return.
        """
        provider_lock = self._active_provider.provider_lock
        try:
            await asyncio.wait_for(provider_lock.acquire(), self._provider_lock_timeout)
            logging.debug("Provider lock acquired, templating.")
            return provider_lock
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError(f"Timeout acquiring provider lock in {self._provider_lock_timeout} seconds.")

    async def render_template_resource(
        self, template_resource_inputs: Dict[str, Any], template_resource_name: str
    ) -> Optional[str]:
        """Render template resource async.

        Args:
            template_resource_name (str): Name of the resource to render, must be registered with active provider.
            template_resource_inputs (Dict[str, Any]): Inputs to template.

        Returns:
            str: templated resource in the form of a string
        """
        active_provider_obj = self._active_provider.provider
        if template_resource_name not in active_provider_obj:
            raise KeyError(f"{template_resource_name} not registered in {active_provider_obj.name} provider.")

        provider_lock = await self.acquire_provider_lock()
        template_render = None
        try:
            resource_obj = active_provider_obj[template_resource_name]
            template = self._active_provider.environment.get_template(resource_obj.template_file_path)
            template_render = await template.render_async(**template_resource_inputs)
        except Exception as e:
            logging.error(f"Error generating template: {e}")
        finally:
            if provider_lock.locked():
                provider_lock.release()
                logging.debug(f"Released {self._active_provider.provider.name} provider lock.")
        return template_render

    async def render_template_resource_to_file(
        self, output_path: str, template_resource_inputs: Dict[str, Any], template_resource_name: str
    ) -> None:
        """Render template and save it to a desired output file.

        Args:
            output_path (str): path to save templated resource.
            template_resource_inputs (Dict[str, Any]): inputs to render template file.
            template_resource_name (str): Name of the resource to template.
        """
        rendered_template_str = await self.render_template_resource(
            template_resource_inputs=template_resource_inputs, template_resource_name=template_resource_name
        )

        if rendered_template_str is not None:
            # verify output file type
            if self.__verify_output_file_type(output_path=output_path, template_resource_name=template_resource_name):
                if os.path.exists(output_path):
                    logging.warning(f"{output_path} exists, overwriting with new templated resource.")
                with open(output_path, "w+") as open_template_file:
                    open_template_file.writelines(rendered_template_str)
        else:
            logging.error("Template failed to render, check logs.")



import os
import tempfile
from typing import Any, Dict

import pytest

from templafirm.core.templater import Templater

from .test_provider import TEST_PROVIDER_META_PATH, BaseProviderTester


@pytest.fixture
def base_test_templater() -> Templater:
    Templater.register_provider("test_provider", BaseProviderTester(TEST_PROVIDER_META_PATH))
    templater = Templater()
    return templater


def test_templater_initialization(base_test_templater: Templater) -> None:
    assert "test_provider" in base_test_templater.list_providers()
    assert "gke" in base_test_templater.list_providers()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_template_str, provider_name, template_inputs, templated_resource",
    [
        ("", "test_provider", {}, "test_empty_resource_template"),
        (
            'resource "test_resource" "test_resource_name" {\n  name                       = "test"\n}',
            "test_provider",
            {},
            "test_no_input_resource_template",
        ),
        (
            'resource "test_resource_input" "test_resource" {\n  input       = "test_input"\n}',
            "test_provider",
            {"resource_name": "test_resource", "input": "test_input"},
            "test_input_resource_template",
        ),
    ],
)
async def test_templater_generation(
    base_test_templater: Templater,
    expected_template_str: str,
    provider_name: str,
    template_inputs: Dict[str, Any],
    templated_resource: str,
) -> None:
    assert provider_name in base_test_templater.list_providers()

    provider_and_env = base_test_templater.return_provider(provider_key=provider_name)
    assert templated_resource in provider_and_env.provider._provider_meta_table.template_mapping.keys()

    base_test_templater.activate_provider("test_provider")
    templated_resource_rendered = await base_test_templater.render_template_resource(
        template_resource_inputs=template_inputs, template_resource_name=templated_resource
    )

    assert templated_resource_rendered == expected_template_str


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_template_str, provider_name, template_inputs, templated_resource",
    [
        ("", "test_provider", {}, "test_empty_resource_template"),
        (
            'resource "test_resource" "test_resource_name" {\n  name                       = "test"\n}',
            "test_provider",
            {},
            "test_no_input_resource_template",
        ),
        (
            'resource "test_resource_input" "test_resource" {\n  input       = "test_input"\n}',
            "test_provider",
            {"resource_name": "test_resource", "input": "test_input"},
            "test_input_resource_template",
        ),
    ],
)
async def test_templater_generation_file(
    base_test_templater: Templater,
    expected_template_str: str,
    provider_name: str,
    template_inputs: Dict[str, Any],
    templated_resource: str,
) -> None:
    assert provider_name in base_test_templater.list_providers()

    provider_and_env = base_test_templater.return_provider(provider_key=provider_name)
    assert templated_resource in provider_and_env.provider._provider_meta_table.template_mapping.keys()

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = os.path.join(tmpdir, "test_template.tf")
        base_test_templater.activate_provider("test_provider")
        await base_test_templater.render_template_resource_to_file(
            output_path=output_path, template_resource_inputs=template_inputs, template_resource_name=templated_resource
        )

        assert os.path.exists(output_path)

        with open(output_path, "r") as open_template_buffer:
            templated_resource = "".join(open_template_buffer.readlines())
        assert templated_resource == expected_template_str
